/*
 * $Id: ProxyRunner.java,v 1.9 2012/01/15 20:59:02 oboehm Exp $
 *
 * Copyright (c) 2011 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 11.11.2011 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.experimental;

import java.lang.annotation.Annotation;
import java.lang.reflect.*;
import java.util.*;

import org.junit.runner.Description;
import org.junit.runner.notification.*;
import org.junit.runners.ParentRunner;
import org.junit.runners.model.*;
import org.slf4j.*;

import patterntesting.runtime.annotation.DelegateTo;
import patterntesting.runtime.junit.SmokeRunner;
import patterntesting.runtime.util.ReflectionHelper;

/**
 * This is a JUnit runner which delegates the call one (or perhaps several)
 * other JUnit runners.
 * </br>
 * At the moment it is experimental and uses hardcoded the internal Runner of
 * JUnit. In the future you should be able to use e.g. Spring's JUnit runner or
 * any other runner.
 * 
 * @author oliver (ob@aosd.de)
 * @since 1.2 (11.11.2011)
 */
public class ProxyRunner extends SmokeRunner {
    
    private static final Logger log = LoggerFactory.getLogger(ProxyRunner.class);
    private ParentRunner<FrameworkMethod> delegateRunner;

    /**
     * Instantiates a new proxy runner.
     *
     * @param testClass the test class
     * @throws InitializationError the initialization error
     */
    public ProxyRunner(final Class<?> testClass) throws InitializationError {
        super(testClass);
    }

    /**
     * Gets the Runner defined by <code>&#064;DelegateTo</code>.
     * <br/>
     * Note: This class is public for testing reasons.
     *
     * @return the delegate runner
     */
    public final ParentRunner<FrameworkMethod> getDelegateRunner() {
        if (this.delegateRunner == null) {
            try {
                this.delegateRunner = createDelegateRunner();
            } catch (InitializationError ie) {
                throw new RuntimeException(ie);
            }
        }
        return this.delegateRunner;
    }
    
    private ParentRunner<FrameworkMethod> createDelegateRunner() throws InitializationError {
        Class<? extends ParentRunner<FrameworkMethod>> runnerClass = getDelegateRunnerClass();
        try {
            Constructor<? extends ParentRunner<FrameworkMethod>> ctor = runnerClass.getDeclaredConstructor(Class.class);
            return ctor.newInstance(this.getTestClass().getJavaClass());
        } catch (SecurityException e) {
            throw new InitializationError(e);
        } catch (NoSuchMethodException e) {
            throw new InitializationError(e);
        } catch (IllegalArgumentException e) {
            throw new InitializationError(e);
        } catch (InstantiationException e) {
            throw new InitializationError(e);
        } catch (IllegalAccessException e) {
            throw new InitializationError(e);
        } catch (InvocationTargetException e) {
            throw new InitializationError(e);
        }
    }

    private Class<? extends ParentRunner<FrameworkMethod>> getDelegateRunnerClass() {
        DelegateTo delegateTo = this.getTestClass().getJavaClass().getAnnotation(DelegateTo.class);
        return delegateTo.value();
    }

    /**
     * Gets the description.
     *
     * @return the description
     * @see org.junit.runner.Runner#getDescription()
     */
    @Override
    public Description getDescription() {
        Description description = this.getDelegateRunner().getDescription();
        Collection<Annotation> annotationList = description.getAnnotations();
        Annotation[] annotations = annotationList.toArray(new Annotation[annotationList.size()]);
        Description filtered = Description.createTestDescription(description.getTestClass(), 
                   description.getDisplayName(), annotations);
        for (Description descr : description.getChildren()) {
            filtered.addChild(descr);
        }
        return filtered;
    }

    /**
     * Here we decide if the whole class is to be run.
     *
     * @param notifier the notifier
     * @see org.junit.runner.Runner#run(org.junit.runner.notification.RunNotifier)
     */
    @Override
    public void run(RunNotifier notifier) {
        //delegateRunner.run(notifier);
        super.run(notifier);
    }
    
//    /**
//     * Describe child. This method will be run only once and will be called
//     * from the constructor to see if there are tests remaining for execution.
//     *
//     * @param method the method
//     * @return the description
//     * @see org.junit.runners.ParentRunner#describeChild(java.lang.Object)
//     */
//    @Override
//    protected Description describeChild(final FrameworkMethod method) {
//        // assert this.delegateRunner != null : "delegateRunner must be initialized!";
//        // TODO replace FrameworkMethod by ProxyFrameworkMethod
//        // return (Description) ReflectionHelper.invokeMethod(delegateRunner, "describeChild", method);
//
//        @Ignore final class Dummy {};
//        Description desc = super.describeChild(method);
//        if (this.getFilter().shouldBeIgnored(desc)) {
//            Annotation[] annotations = method.getAnnotations();
//            Annotation[] withIgnored = new Annotation[annotations.length + 1];
//            System.arraycopy(annotations, 0, withIgnored, 0, annotations.length);
//            withIgnored[withIgnored.length-1] = Dummy.class.getAnnotation(Ignore.class);
//            desc = Description.createTestDescription(getTestClass().getJavaClass(), method.getName(), withIgnored);
//        }
//        return desc;
//    }

    /**
     * Gets the children.
     *
     * @return the children
     * @see org.junit.runners.ParentRunner#getChildren()
     */
    @SuppressWarnings("unchecked")
    @Override
    protected List<FrameworkMethod> getChildren() {
        // TODO: put a Proxy around each FrameworkMethod
        List<FrameworkMethod> methods = (List<FrameworkMethod>) ReflectionHelper.invokeMethod(this.getDelegateRunner(), "getChildren");
        return methods;
    }

    /**
     * Run child. This method may be not called (e.g. it is not called by 
     * SpringJUnit4ClassRunner) but must be implemented because of the abstract
     * superclass (ParentRunner).
     *
     * @param method the method
     * @param notifier the notifier
     * @see org.junit.runners.ParentRunner#runChild(java.lang.Object, org.junit.runner.notification.RunNotifier)
     */
    @Override
    protected void runChild(final FrameworkMethod method, final RunNotifier notifier) {
        Description description = describeChild(method);
        try {
            if (shouldBeIgnored(method)) {
                notifier.fireTestIgnored(description);
                return;
            }
        } catch (IllegalArgumentException iae) {
            notifier.fireTestAssumptionFailed(new Failure(description, iae));
            notifier.fireTestFinished(description);
            return;
        }
        long startTime = System.currentTimeMillis();
        ReflectionHelper.invokeMethod(delegateRunner, "runChild", method, notifier);
        logMethod(method, System.currentTimeMillis() - startTime);
    }

    private static void logMethod(final FrameworkMethod method, final long time) {
        log.info("{}.{} (" + time + " ms)", method.getMethod().getDeclaringClass().getName(),
                method.getName());
    }

}

